/*
 * Copyright 2017 Advanced Micro Devices, Inc.
 * Copyright 2020-2024 Hewlett Packard Enterprise Development LP
 * Copyright 2004-2019 Cray Inc.
 * Other additional copyright holders may be indicated within.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// LocaleModel.chpl
//
// This provides a flat locale model architectural description.  The
// locales contain memory and a multi-core processor with homogeneous
// cores, and we ignore any affinity (NUMA effects) between the
// processor cores and the memory.  This architectural description is
// backward compatible with the architecture implicitly provided by
// releases 1.6 and preceding.
//
module LocaleModel {

  public use LocaleModelHelpAPU;
  public use LocaleModelHelpMem;

  use IO;

  //
  // The task layer calls these to convert between full sublocales and
  // execution sublocales.  Full sublocales may contain more information
  // in some locale models, but not in this one.
  //
  export
  proc chpl_localeModel_sublocToExecutionSubloc(full_subloc:chpl_sublocID_t)
  {
    return full_subloc;  // execution sublocale is same as full sublocale
  }

  export
  proc chpl_localeModel_sublocMerge(full_subloc:chpl_sublocID_t,
                                    execution_subloc:chpl_sublocID_t)
  {
    return execution_subloc;  // no info needed from full sublocale
  }

  class CPULocale : AbstractLocaleModel, writeSerializable {
    const sid: chpl_sublocID_t;
    const name_: string;

    override proc chpl_id() do return (parent:LocaleModel)._node_id; // top-level node id
    override proc chpl_localeid() {
      return chpl_buildLocaleID((parent:LocaleModel)._node_id:chpl_nodeID_t,
                                sid);
    }
    override proc chpl_name() do return name_;

    proc init() {
    }

    proc init(_sid, _parent) {
      super.init(new locale(_parent));
      sid = _sid;
      name_ = _parent.chpl_name() + "-CPU" + sid:string;
    }

    override proc serialize(writer, ref serializer) throws {
      parent.serialize(writer, serializer);
      writer.write('.'+name_);
    }

    override proc _getChildCount(): int { return 0; }

    iter getChildIndices() : int {
      halt("No children to iterate over.");
    }
    proc addChild(loc:locale) {
      halt("Cannot add children to this locale type.");
    }
    override proc _getChild(idx:int) : locale {
      halt("requesting a child from a CPULocale locale");
    }
  }

  class GPULocale : AbstractLocaleModel {
    const sid: chpl_sublocID_t;
    const name_: string;

    override proc chpl_id() do return (parent:LocaleModel)._node_id; // top-level node id
    override proc chpl_localeid() {
      return chpl_buildLocaleID((parent:LocaleModel)._node_id:chpl_nodeID_t,
                                sid);
    }
    override proc chpl_name() do return name_;

    proc init() {
    }
    proc deinit() {
     extern proc hsa_shutdown(): void ;

     // Comment out this until runtime support for HSA is added

     //     hsa_shutdown();
    }


    proc init(_sid, _parent) {
      super.init(new locale(_parent));
      sid = _sid;
      name_ = _parent.chpl_name() + "-GPU" + sid:string;
    }

    override proc serialize(writer, ref serializer) throws {
      parent.serialize(writer, serializer);
      writer.write('.'+name_);
    }

    override proc _getChildCount(): int { return 0; }
    iter getChildIndices() : int {
      halt("No children to iterate over.");
    }
    proc addChild(loc:locale) {
      halt("Cannot add children to this locale type.");
    }
    override proc _getChild(idx:int) : locale {
      halt("requesting a child from a GPULocale locale");
    }
  }

  const chpl_emptyLocaleSpace: domain(1) = {1..0};
  pragma "unsafe"
  const chpl_emptyLocales: [chpl_emptyLocaleSpace] locale;

  //
  // The node model
  //
  class LocaleModel : AbstractLocaleModel {
    const _node_id : int;
    var local_name : string; // should never be modified after first assignment

    var numSublocales: int; // should never be modified after first assignment
    var GPU: unmanaged GPULocale?;
    var CPU: unmanaged CPULocale?;

    // This constructor must be invoked "on" the node
    // that it is intended to represent.  This trick is used
    // to establish the equivalence the "locale" field of the locale object
    // and the node ID portion of any wide pointer referring to it.
    proc init() {
      if rootLocaleInitialized {
        halt("Cannot create additional LocaleModel instances");
      }
      _node_id = chpl_nodeID: int;

      init this;

      setup();
    }

    proc init(parent_loc : locale) {
      if rootLocaleInitialized {
        halt("Cannot create additional LocaleModel instances");
      }
      super.init(parent_loc);

      _node_id = chpl_nodeID: int;

      init this;

      setup();
    }

    override proc chpl_id() do return _node_id;     // top-level locale (node) number
    override proc chpl_localeid() {
      return chpl_buildLocaleID(_node_id:chpl_nodeID_t, c_sublocid_none);
    }
    override proc chpl_name() do return local_name;

    proc getChildSpace(): domain(1) {
      halt("error!");
    }

    override proc _getChildCount() do return 0;

    iter getChildIndices() : int {
      for idx in {0..#numSublocales} do // chpl_emptyLocaleSpace do
        yield idx;
    }

    override proc _getChild(idx:int) : locale {
      if idx == 1
        then return new locale(GPU!);
      else
        return new locale(CPU!);
    }

    iter getChildren() : locale  {
      halt("in here");
      for idx in {0..#numSublocales} {
        if idx == 1
          then yield GPU;
        else
          yield CPU;
      }
    }

    proc getChildArray(): chpl_emptyLocales.type {
      halt ("in get child Array");
    }


    //------------------------------------------------------------------------{
    //- Implementation (private)
    //-
    proc setup() {
      helpSetupLocaleAPU(this, local_name, numSublocales, CPULocale, GPULocale);
    }
    //------------------------------------------------------------------------}

    proc deinit() {
      if (debugAPULocale) {
        chpl_debug_writeln("** Destructing CPU/GPU locales and shutting down HSA.");
      }
      delete _to_unmanaged(CPU);
      delete _to_unmanaged(GPU);
    }
 }

  //
  // An instance of this class is the default contents 'rootLocale'.
  //
  // In the current implementation a platform-specific architectural description
  // may overwrite this instance or any of its children to establish a more customized
  // representation of the system resources.
  //
  class RootLocale : AbstractRootLocale {

    const myLocaleSpace: domain(1) = {0..numLocales-1};
    pragma "unsafe"
    var myLocales: [myLocaleSpace] locale;

    proc init() {
      super.init(nilLocale);
      nPUsPhysAcc = 0;
      nPUsPhysAll = 0;
      nPUsLogAcc = 0;
      nPUsLogAll = 0;
      maxTaskPar = 0;
    }

    // The setup() function must use chpl_initOnLocales() to iterate (in
    // parallel) over the locales to set up the LocaleModel object.
    // In addition, the initial 'here' must be set.
    proc setup() {
      helpSetupRootLocaleAPU(this);
    }

    // Has to be globally unique and not equal to a node ID.
    // We return numLocales for now, since we expect nodes to be
    // numbered less than this.
    // -1 is used in the abstract locale class to specify an invalid node ID.
    override proc chpl_id() do return numLocales;
    override proc chpl_localeid() {
      return chpl_buildLocaleID(numLocales:chpl_nodeID_t, c_sublocid_none);
    }
    override proc chpl_name() do return local_name();
    proc local_name() do return "rootLocale";

    override proc serialize(writer, ref serializer) throws {
      writer.write(name);
    }

    override proc _getChildCount() do return this.myLocaleSpace.size;

    proc getChildSpace() do return this.myLocaleSpace;

    iter getChildIndices() : int {
      for idx in this.myLocaleSpace do
        yield idx;
    }

    override proc _getChild(idx:int) do return this.myLocales[idx];

    iter getChildren() : locale  {
      for loc in this.myLocales do
        yield loc;
    }

    override proc getDefaultLocaleSpace() const ref do return this.myLocaleSpace;
    override proc getDefaultLocaleArray() const ref do return myLocales;

    override proc localeIDtoLocale(id : chpl_localeID_t) {
      const node = chpl_nodeFromLocaleID(id);
      const subloc = chpl_sublocFromLocaleID(id);
      if (subloc == c_sublocid_none) || (subloc == c_sublocid_none) then
        return (myLocales[node:int]):locale;
      else
        return (myLocales[node:int]._getChild(subloc:int)):locale;
    }

    proc deinit() {
      for loc in myLocales {
        on loc {
          rootLocaleInitialized = false;
        }
      }
    }
  }
}
